import itertools
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import NoReturn, Optional, Union

from tree_helper_api import data_tree_walker as walk
from unit_api import (
    format_date,
    format_units,
    get_first_short_title_from_units,
    load_data,
    trim_text,
)

from .statute_identity import StatuteID, get_member

MAX_TITLE_LENGTH = 1000  # creation of statute title types


@dataclass
class StatuteItem:
    """Each generated item contains attributes sourced from metadata yaml files"""

    category: str
    idx: str

    def __post_init__(self):
        self.category: str = self.category.upper()
        self.idx: str = self.idx.upper()
        self.member: StatuteID = get_member(self.category)
        self.folder: Path = self.member.data_source_folder(self.idx)
        self.data: dict = self.get_data
        self.specified_date: date = format_date(self.data["date"])
        self.units: list[dict] = self.data["units"]
        self.effects: list[str] = self.data.get("effects", None)

    @property
    def get_data(self) -> Union[NoReturn, dict]:
        if not (raw_data := load_data(self.folder)):
            raise Exception(f"No data from {self.folder}")
        for key in ["item", "law_title", "date", "units"]:
            if key not in raw_data:
                raise Exception(f"Missing {key}")
        format_units(raw_data["units"])
        return raw_data

    @property
    def effective_actions(self) -> list[dict]:
        return list(itertools.chain(*walk(self.data, "effects")))

    """Format titles based on category, idx, titles in data"""

    @property
    def titles(self) -> dict:
        return {
            "short": self.short_title,
            "official": self.official_title,
            "serial": self.serial_title,
            "aliases": self.data.get("aliases", []),
        }

    @property
    def short_title(self) -> Optional[str]:
        return get_first_short_title_from_units({"units": self.units})

    @property
    def official_title(self) -> Optional[str]:
        raw = self.data.get("law_title", None) or self.data.get("item", None)
        return trim_text(raw, MAX_TITLE_LENGTH) if raw else None

    @property
    def serial_title(self) -> Optional[str]:  # previously just "title"
        raw = self.member.make_title(self.idx)
        return trim_text(raw, MAX_TITLE_LENGTH) if raw else None
