from enum import Enum, auto, unique
from pathlib import Path
from typing import Optional

from statute_api import STATUTES_PATH

_NUMBER = "No."
_BILANG = "Blg."
_DATED = "dated"
_DASHED = "-"


@unique
class TitleConventions(Enum):
    SERIAL = auto()  # e.g. Republic Act No. 386
    SHORT = auto()  # e.g. Civil Code of the Philippines
    OFFICIAL = auto()  # e.g. An Act to Ordain and Institute the...
    ALIAS = auto()  # e.g. New Civil Code


def get_title_choices() -> list[str]:
    return [name for name, member in TitleConventions.__members__.items()]


@unique
class StatuteID(Enum):
    """
    Decoding & encoding of statutory identifiers. The `names` should be shortcut folder source of raw statutory yaml files. The `values` are the category names.

    The serialized value constitutes, e.g. `1987 Constitution` or the `Spanish Civil Code`. The serialized value can be broken down into a `name` and `identifier`. In the case of the `1987 Constitution`, the name would be "CONST" and the identifier: "1987"
    The names are uniform to enable bucketing similar statutory files together. The full serialized form, e.g. `Republic Act No. 1241`, is what is utilized to reference MatchedTitleProvisions.
    """

    # Note that `Republic Act` and `Commonwealth Act` must precede Act
    # name = (value[0], value[1])
    RA = ("Republic Act", _NUMBER)
    CA = ("Commonwealth Act", _NUMBER)
    ACT = ("Act", _NUMBER)
    CONST = ("Constitution", None)
    SPAIN = ("Spanish", None)
    BP = ("Batas Pambansa", _BILANG)
    PD = ("Presidential Decree", _NUMBER)
    EO = ("Executive Order", _NUMBER)
    VETO = ("Veto Message", _DASHED)
    RULE_ROC = ("Rules of Court", None)
    RULE_BM = ("Bar Matter", _NUMBER)
    RULE_AM = ("Administrative Matter", _NUMBER)
    RULE_RESO = ("Resolution of the Court En Banc", _DATED)

    @property
    def parts(self) -> str:
        """Remove `None` values when joining the tuple value"""
        return " ".join(elem for elem in self.value if elem)

    @property
    def meta_folder(self) -> Path:
        """Assuming a base path `statutes` folder, determine if a subfolder therein exists which match the member's name"""
        return STATUTES_PATH.joinpath(self.name)

    def data_source_folder(self, idx: str) -> Optional[Path]:
        """Get the associated folder, subfolder of an existing `meta_folder`"""
        return (
            subfolder
            if self.meta_folder.exists()
            and (subfolder := self.meta_folder.joinpath(idx))
            and subfolder.exists()
            and subfolder.is_dir()
            else None
        )

    def get_spanish_id_code(self, text: str):
        """Legacy docs don't have serialized identifiers"""
        remainder = text.removeprefix("Spanish ")
        if "civil" in remainder.lower():
            return "CIVIL"
        elif "commerce" in remainder.lower():
            return "COMMERCE"
        elif "penal" in remainder.lower():
            return "PENAL"

    def get_idx(self, txt: str) -> str:
        """Given text e.g. `Spanish Civil Code` or `Executive Order No. 111`, get the serial number"""
        if self.name == "SPAIN" and (code := self.get_spanish_id_code(txt)):
            return code  # special case
        return txt.replace(self.parts, "").strip()  # regular

    def search_pair(self, txt: str) -> Optional[tuple[str, str]]:
        """Return shortcut tuple of member name and identifier, if found."""
        return (self.name, self.get_idx(txt)) if self.value[0] in txt else None

    def make_title(self, idx: str) -> str:
        """Return full title; notice inverted order for Rules of Court, Constitution"""
        if self.name in ["RULE_ROC", "CONST"]:
            return f"{idx} {self.parts}"  # e.g. 1987 Constitution
        elif self.name == "SPAIN":
            if code := self.get_spanish_id_code(f"Spanish {idx}"):
                if code == "CIVIL" or code == "PENAL":
                    return f"Spanish {code.title()} Code"
                elif code == "COMMERCE":
                    return "Spanish Code of Commerce"
        return f"{self.parts} {idx}"  # e.g. Republic Act No. 1231


def get_statute_choices() -> list[tuple[str, str]]:
    return [
        (name, member.value[0])
        for name, member in StatuteID.__members__.items()
    ]


def get_member(query: str) -> Optional[StatuteID]:
    for name, member in StatuteID.__members__.items():
        if query == name:
            return member
    return None


def get_category_idx(text: str) -> Optional[tuple[str, str]]:
    """Matching category ("RA") and identifier ("8908") from statutory text, e.g. Republic Act No. 8908"""
    for member in StatuteID:
        if pair := member.search_pair(text):
            return pair
    return None
