import re
from typing import Match, Optional


class ProvisionLabel:
    paragraph = r"""
        \b
        (sub-?)?pars?
        (
            \.
            |agraphs?
        )?
        \b
    """

    article = r"""
        \b
        (A|a)rts?
        (
            \.
            |(icles?\b)
        )?
    """

    article_capped = r"""
        \b
        ARTS?\.?
        \b
    """

    section = r"""
        \b
        (S|s)
        ec
        (
            \.
            |(tions?)\b
        )?

    """

    section_capped = r"""
        \b
        SEC(TION)?S?\.?
        \b
    """

    chapter = r"""
        \b
        (C|c)h
        (
            \.
            |apters?
        )?
        \b
    """

    book = r"""
        \b
        (B|b)
        (
            (ook)
            |(k.?)
        )
        \b
    """

    title = r"""
        (?<![Cc]ertificate\sof\s) # without s Original Certificate of Title
        (?<![Cc]ertificates\sof\s) # with s # Transfer Certificates of Title
        \b
        (T|t)it
        (
            \.
            |le
        )?
        \b
    """

    rule = r"""
        \bRule\b
    """

    canon = r"""
        \bCanon\b
    """

    section_symbol = r"""
        §§?
    """

    formula = rf"""
        (
            ({rule}\s+)|
            ({canon}\s+)|
            ({paragraph}\s+)|
            ({article}\s+)|
            ({article_capped}\s+)|
            ({section}\s+)|
            ({section_capped}\s+)|
            ({section_symbol}\s*)| # enables `§14` and `§ 14` using s*
            ({title}\s+)|
            ({chapter}\s+)|
            ({book}\s+)
        )
    """


class ProvisionSubject:
    just_digits = r"""
        (
            \d+
        )
    """

    simple_roman = r"""
        (
            [IXV]+\b
        ) #\b ensures that in cases like "Section 6 of PD 902-A provides:"SECTION 6. In " SECTION 6. excludes an additional "I"
    """

    simple_letters = r"""
        (
            \- # ensures that -A, -B, -C is used; prevents matching of `B.P.`
            [ABC]
            \b
        )
    """

    bracketed_character = r"""
        (
            \(
            \w{1,3}
            \)
        )
    """

    connector = r"""
        (
            and|
            of|
            \,|
            \.|
            \s|
            \-
        )
    """

    formula = rf"""
        ( # must start with an initial object
            {just_digits}|
            {simple_roman}|
            {bracketed_character}
        ) # may end with multiple objects
        (
            {simple_letters}|
            {simple_roman}|
            {just_digits}|
            {bracketed_character}|
            {connector}
        )*
    """


class Provision:

    formula = rf"""
        (
            {ProvisionLabel().formula}
            {ProvisionSubject().formula}
        )+
    """


class ProvisionMatcher:
    def __init__(self, text) -> None:
        provision_regex = Provision().formula
        provision_pattern = re.compile(provision_regex, re.X)
        self.results = list(provision_pattern.finditer(text))


def get_label(text) -> Optional[Match]:
    label_regex = ProvisionLabel().formula
    label_pattern = re.compile(label_regex, re.X)
    return label_pattern.search(text)


def isolated(text) -> Optional[str]:
    """Removes the label e.g. Art., Section from the query_text"""
    if label_matched := get_label(text):
        return re.sub(label_matched.group(), "", text).strip()
    return None
