import re
from collections import deque, namedtuple
from dataclasses import dataclass
from enum import Enum
from typing import Match, NamedTuple, Optional, Pattern

from ..statute_identity import code_translate, get_member, rule_translate
from .formula import TITLE, UNIT, isolated


class MatchedTitleProvision(NamedTuple):
    """Based on `text` representing a statute's identity, the following breakdown of fields can be detected:
    1. a `title` (the coded category of the statute), e.g. 'RA';
    2. a `num` (the unique identifier of the category), e.g. '386',
    3. a `provision` of the statute, e.g.'art. 2'"
    """

    text: str
    title: Optional[str]
    num: Optional[str]
    provision: Optional[str]

    @property
    def is_statutory(self) -> bool:
        "Contains both a title and num field"
        return True if self.title and self.num else False

    @property
    def cat_idx(self) -> Optional[tuple[str, str]]:
        "Tuple containing the title and num field"
        return (self.title, self.num) if self.title and self.num else None

    @property
    def statute_id(self) -> Optional[str]:
        "Combined identified title and num based on a StatuteID (enum)"
        if self.title and self.num and (member := get_member(self.title)):
            return member.make_title(self.num)
        return None

    @property
    def isolated_provision(self) -> Optional[str]:
        return isolated(self.provision) if self.provision else None


def compiler(text_to_query: str) -> Optional[list[MatchedTitleProvision]]:
    """
    Compile list of MatchedTitleProvisions
    >>> sample = "There is no question that Section 2 of Presidential Decree No. 1474-B is inconsistent with Section 62 of Republic Act No. 3844.; Petitioner’s case was decided under P.D. No. 971, as amended by P.D. No. 1707."
    >>> compiler(sample)
    [MatchedTitleProvision(text='Section 2 of Presidential Decree No. 1474-B', title='PD', num='1474-B', provision='Section 2'),
    MatchedTitleProvision(text='Section 62 of Republic Act No. 3844', title='RA', num='3844', provision='Section 62'),
    MatchedTitleProvision(text='P.D. No. 971', title='PD', num='971', provision=None),
    MatchedTitleProvision(text='P.D. No. 1707', title='PD', num='1707', provision=None)]
    """
    return [
        matched_title_provision
        for text_string in merge_linked_spans(text_to_query)
        if (matched_title_provision := make_title_provision(text_string))
    ]


def merge_linked_spans(input: str):
    """
    Combine spans if they are "next" to each other and are of different item classes, i.e. provision and law.
    >>> sample = "There is no question that Section 2 of Presidential Decree No. 1474-B is inconsistent with Section 62 of Republic Act No. 3844.; Petitioner’s case was decided under P.D. No. 971, as amended by P.D. No. 1707."
    >>> merge_linked_spans(sample)
    ['Section 2 of Presidential Decree No. 1474-B',
    'Section 62 of Republic Act No. 3844',
    'P.D. No. 971',
    'P.D. No. 1707']
    """

    def item_class(text: str):
        Item = namedtuple("Item", ["text", "category"])
        if Context(text, Style.LAW.value).match:
            return Item(text, "title")
        elif Context(text, Style.PROV.value).match:
            return Item(text, "unit")

    strings: list[str] = []

    if not (items := spans(input)):
        return []

    # instantiate queue, assumes more than one item
    q = deque(items)

    # exhaust queue
    while q:  #

        # get items
        prior = q.popleft()
        former = item_class(prior.group())
        try:
            now = q.popleft()
            latter = item_class(now.group())
        except:
            strings.append(prior.group())
            break

        # determine whether or not to combine
        if former.category != latter.category:

            # obvious connection, always combine
            if (distance := prior.end() - now.start()) == 0:
                item = prior.group() + now.group()
                strings.append(item)
                continue

            # probable connection, consider combination
            elif abs(distance) <= 5:
                if "." not in (mid := input[prior.end() : now.start()]):
                    if ";" not in (mid):
                        item = prior.group() + mid + now.group()
                        strings.append(item)
                        continue

        # no connection, do not combine, add
        strings.append(prior.group())
        q.appendleft(now)

    return strings


class Style(Enum):
    """
    PROVLAW format: "Const., Art. III"
    LAWPROV format: "CIVIL CODE, Articles 1547(1), 1548-1560"
    """

    LAW = re.compile(TITLE, re.X)  # CONST, RA 354, Republic Act No. (RA) 354
    PROV = re.compile(UNIT, re.X)  # Sec. 1, Art. 2350 (a), Bk V, Title XV


@dataclass
class Context:
    """
    >>> sample = "prior to its amendment by Republic Act (RA) No. 8424, otherwise known as the Tax Reform Act of 1997; Section 533 of Rep. Act 7160 reads in part:"
    >>> Context(sample, Style.PROVLAW.value).match
    <re.Match object; span=(101, 129), match='Section 533 of Rep. Act 7160'>
    >>> Context(sample, Style.PROVLAW.value).matches
    [<re.Match object; span=(101, 129), match='Section 533 of Rep. Act 7160'>]
    """

    input: str  # text to contextualize
    pattern: Pattern  # a possible Style

    @property
    def matches(self) -> list[Match]:
        return list(self.pattern.finditer(self.input))

    @property
    def match(self) -> Optional[Match]:
        return self.pattern.search(self.input)


def spans(input: str) -> list[Match]:
    """
    Get possible statute text ranges from text, ordered by their character position in the text
    >>> sample = "There is no question that Section 2 of Presidential Decree No. 1474-B is inconsistent with Section 62 of Republic Act No. 3844.; Petitioner’s case was decided under P.D. No. 971, as amended by P.D. No. 1707."
    >>> spans(sample)
    [<re.Match object; span=(26, 39), match='Section 2 of '>,
    <re.Match object; span=(39, 69), match='Presidential Decree No. 1474-B'>,
    <re.Match object; span=(91, 105), match='Section 62 of '>,
    <re.Match object; span=(105, 126), match='Republic Act No. 3844'>,
    <re.Match object; span=(165, 177), match='P.D. No. 971'>,
    <re.Match object; span=(193, 206), match='P.D. No. 1707'>]
    """
    units = [
        *Context(input, Style.LAW.value).matches,
        *Context(input, Style.PROV.value).matches,
    ]
    return sorted(units, key=lambda span: span.start())


def make_title_provision(item: str) -> Optional[MatchedTitleProvision]:
    """
    Create a statute entity based on a preprocessed / filtered item
    >>> sample = "Republic Act (RA) No. 8424"
    >>> make_title_provision(sample)
    MatchedTitleProvision(text='Republic Act (RA) No. 8424', title='RA', num='8424', provision=None)
    """

    title = Context(item, Style.LAW.value).match
    proviso = (
        cull(provision.group())
        if (provision := Context(item, Style.PROV.value).match)
        else None
    )

    if title:
        return MatchedTitleProvision._make([item, *categorize(title), proviso])

    elif proviso:
        return MatchedTitleProvision._make([item, None, None, proviso])

    return None


def categorize(match: Match) -> tuple[Optional[str], Optional[str]]:
    """
    Expected: match object from bigger string
    string = "There is no question that Section 2 of Presidential Decree No. 1474-B is inconsistent with Section 62 of Republic Act No. 3844.; Petitioner’s case was decided under P.D. No. 971, as amended by P.D. No. 1707."
    >>> sample = filtered(string)[0]
    >>> sample
    <re.Match object; span=(26, 69), match='Section 2 of Presidential Decree No. 1474-B'>
    >>> categorize(sample)
    ('PD', '1474-B')
    """
    if match.group("const_doc"):
        return (
            "CONST",
            year_found.strip()
            if (year_found := match.group("const_year"))
            else None,
        )

    elif match.group("legal_code"):
        codal = cull(match.group("legal_code"))
        if res := code_translate(codal):
            return res[0], res[1]
        return "CODE", codal

    elif match.group("rules_court"):
        rule_text = cull(match.group("rules_court"))
        if res := rule_translate(rule_text):
            return res[0], res[1]
        return "RULE_ROC", rule_text

    elif match.group("rule_am"):
        return "RULE_AM", cull(match.group("AM_digit"))

    elif match.group("rule_bm"):
        return "RULE_BM", cull(match.group("BM_digit"))

    elif match.group("rule_reso"):
        return "RULE_RESO", cull(match.group("RESO_digit"))

    elif match.group("republic_act"):
        return "RA", cull(match.group("RA_digit"))

    elif match.group("presidential_decree"):
        return "PD", cull(match.group("PD_digit"))

    elif match.group("batas_pambansa"):
        return "BP", cull(match.group("BP_digit"))

    elif match.group("executive_order"):
        return "EO", cull(match.group("EO_digit"))

    elif match.group("commonwealth_act"):
        return "CA", cull(match.group("CA_digit"))

    elif match.group("legacy_act"):
        return "ACT", cull(match.group("ACT_digit"))

    return None, None


def cull(text: str) -> str:
    for i in ["the", "of", ","]:
        text = text.removesuffix(i).strip()
    return text
